"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spread = (this && this.__spread) || function () {
    for (var ar = [], i = 0; i < arguments.length; i++) ar = ar.concat(__read(arguments[i]));
    return ar;
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
Object.defineProperty(exports, "__esModule", { value: true });
var debugGenerator = require("debug");
var path = require("path");
var fs = require("fs-extra");
var RSVP = require("rsvp");
var mkdirp = require("mkdirp");
var BroccoliPlugin = require("broccoli-plugin");
var glob = require("glob");
var FSTree = require("fs-tree-diff");
var walkSync = require("walk-sync");
var queue = require("async-promise-queue");
var ensureSymlink = require("ensure-symlink");
var nodeSass = require("node-sass");
var copyObject = require("lodash.clonedeep");
var MergeTrees = require("broccoli-merge-trees");
var chained_emitter_1 = require("chained-emitter");
var DiskCache = require("sync-disk-cache");
var heimdall = require("heimdalljs");
var fs_1 = require("fs");
var concurrency_1 = require("./concurrency");
var concurrency = RSVP.resolve(concurrency_1.determineOptimalConcurrency());
var FSTreeFromEntries = FSTree.fromEntries;
var debug = debugGenerator("broccoli-eyeglass");
var hotCacheDebug = debugGenerator("broccoli-eyeglass:hot-cache");
var concurrencyDebug = debug.extend("concurrency");
var sass;
var renderSass;
function initSass() {
    if (!sass) {
        sass = nodeSass;
        // Standard async rendering for node-sass.
        renderSass = RSVP.denodeify(sass.render);
    }
}
function absolutizeEntries(entries) {
    // We make everything absolute because relative path comparisons don't work for us.
    entries.forEach(function (entry) {
        // TODO support windows paths
        entry.relativePath = path.join(entry.basePath, entry.relativePath);
        entry.basePath = "/";
    });
}
function shouldPersist(env, persist) {
    var result;
    if (env.CI) {
        result = env.FORCE_PERSISTENCE_IN_CI;
    }
    else {
        result = persist;
    }
    return !!result;
}
var Entry = /** @class */ (function () {
    function Entry(path) {
        var stats = fs.statSync(path);
        this.relativePath = path;
        this.basePath = "/";
        this.mode = stats.mode;
        this.size = stats.size;
        this.mtime = stats.mtime;
    }
    Entry.prototype.isDirectory = function () {
        return false;
    };
    return Entry;
}());
function unique(array) {
    return new (Array.bind.apply(Array, __spread([void 0], new Set(array))))();
}
// This promise runs sass synchronously
function renderSassSync(options) {
    return RSVP.resolve(sass.renderSync(options));
}
function removePathPrefix(prefix, fileNames) {
    if (prefix[prefix.length - 1] !== path.sep) {
        prefix = prefix + path.sep;
    }
    var newFileNames = new Array();
    for (var i = 0; i < fileNames.length; i++) {
        if (fileNames[i].indexOf(prefix) === 0) {
            newFileNames[i] = fileNames[i].substring(prefix.length);
        }
        else {
            newFileNames[i] = fileNames[i];
        }
    }
    return newFileNames;
}
// sassFile: The sass file being compiled
// cssFile: The default css file location.
// cb: This callback must be invoked once for each time you want to compile the
//     sass file. It must be called synchronously. You can change the output
//     filename and options passed to it.
var defaultOptionsGenerator = function (_sassFile, cssFile, options, cb) { return cb(cssFile, options); };
// Support for older versions of node.
function parsePath(pathname) {
    if (path.parse) {
        return path.parse(pathname);
    }
    else {
        var parsed = {
            root: "",
            name: "",
            dir: path.dirname(pathname),
            base: path.basename(pathname),
            ext: path.extname(pathname),
        };
        parsed.name = parsed.base.substring(0, parsed.base.length - parsed.ext.length);
        return parsed;
    }
}
function formatPath(parsed) {
    if (path.format) {
        return path.format(parsed);
    }
    else {
        return path.join(parsed.dir, parsed.name + parsed.ext);
    }
}
function forbidNodeSassOption(options, property) {
    if (options[property]) {
        throw new Error("The node-sass option '" + property + "' cannot be set explicitly.");
    }
}
/* write data to cachedFile, and symlink outputFile to that
 *
 * @argument cachedFile - the file to write the data to
 * @argument outputFile - where to write the symlink
 * @argument data - the data to write
 */
function writeDataToFile(cachedFile, outputFile, data) {
    mkdirp.sync(path.dirname(cachedFile));
    fs.writeFileSync(cachedFile, data);
    mkdirp.sync(path.dirname(outputFile));
    ensureSymlink(cachedFile, outputFile);
}
// This Sass compiler has a different philosophy than the default one
// that comes with broccoli. It is directory based instead of being file
// based and can merge several input trees into a single output tree
// instead of using the trees as a proxy for includePaths. It has error
// handling, verbose logging and will pass through any options that
// it doesn't own. It uses the node-sass async api via promises.
// It allows several css files to be compiled from a single sass file
// by customizing the options and output file names.
//
// You can emit the following events from custom functions that are invoked during compilation:
// * compiler.events.emit("dependency", absolutePath);
//     marks the file as a dependency for the Sass file being compiled so that future
//     compiles will invalidate the cache if that file changes.
// * compiler.events.emit("additional-output", absolutePathToOutput, httpPathToOutput, absolutePathToSource);
//     marks the file as an additional output for the Sass file being compiled so that future
//     cached compiles will be able to install or remove them as needed in conjunction with the
//     sass file. Note: the source will not be considered a dependency unless the "dependency" event
//     is also emitted.
//
// You can subscribe to the following events:
//
//   * compiler.events.on("compiling", function(details) { });
//       prepare for a compilation to occur with these options. The options
//       are a unique copy for this compilation. E.g. This can be used to
//       prime a cache in the options for the compilation.
//   * compiler.events.on("compiled", (details, result) => { });
//       receive notification of a successful compilation.
//   * compiler.events.on("failed", (details, error) => { });
//       receive notification of a compilation failure.
//   * compiler.events.on("stale-external-output", (outputFile) => {})
//       receive a notification that a file outside of this broccoli output
//       tree might need to be removed because the only known sass file
//       in this tree to output it has been deleted.
//   * compiler.events.on("cached-asset", (absolutePathToSource, httpPathToOutput) => {})
//       receive a notification that an additional asset that was created
//       when the caller fired "additional-output" (see above) needs to be
//       restored because the sass file that produced it was retrieved from
//       cache. This is only invoked when the asset was created outside of the
//       broccoli tree for this addon. If the asset was in the tree, it will
//       be automatically recreated from the cache.
//
//   For all these events, a compilation details object is passed of the
//   following form:
//
//   {
//     srcPath: /* Source directory containing the sassFilename */,
//     sassFilename: /* Sass file relative to the srcPath */,
//     fullSassFilename: /* Fully expanded and resolved sass filename. */,
//     destDir: /* The location cssfiles are being written (a tmp dir) */,
//     cssFilename: /* The css filename relation to the output directory */,
//     fullCssFilename: /* Fully expanded and resolved css filename */,
//     options: /* The options used for this compile */
//   };
//
var BroccoliSassCompiler = /** @class */ (function (_super) {
    __extends(BroccoliSassCompiler, _super);
    function BroccoliSassCompiler(inputTree, options) {
        var _this = this;
        if (Array.isArray(inputTree)) {
            if (inputTree.length > 1) {
                // eslint-disable-next-line no-console
                console.warn("Support for passing several trees to BroccoliSassCompiler has been removed.\n" +
                    "Passing the trees to broccoli-merge-trees with the overwrite option set,\n" +
                    "but you should do this yourself if you need to compile CSS files from them\n" +
                    "or use the node-sass includePaths option if you need to import from them.");
                inputTree = new MergeTrees(inputTree, { overwrite: true, annotation: "Sass Trees" });
            }
            else {
                inputTree = inputTree[0];
            }
        }
        options = options || {};
        options.persistentOutput = true;
        _this = _super.call(this, [inputTree], options) || this;
        _this.buildCount = 0;
        _this.events = new chained_emitter_1.EventEmitter();
        _this.currentTree = null;
        _this.dependencies = {};
        _this.outputs = {};
        _this.outputURLs = {};
        _this.sessionCache = options.sessionCache;
        delete options.sessionCache;
        _this.buildCache = _this.sessionCache || new Map();
        if (shouldPersist(process.env, !!options.persistentCache)) {
            _this.persistentCache = new DiskCache(options.persistentCache);
        }
        _this.persistentCacheDebug = debugGenerator("broccoli-eyeglass:persistent-cache:" + (options.persistentCache || 'disabled'));
        _this.treeName = options.annotation;
        delete options.annotation;
        _this.cssDir = options.cssDir;
        delete options.cssDir;
        _this.sassDir = options.sassDir;
        delete options.sassDir;
        _this.optionsGenerator = options.optionsGenerator || defaultOptionsGenerator;
        delete options.optionsGenerator;
        _this.fullException = options.fullException || false;
        delete options.fullException;
        _this.verbose = options.verbose || debugGenerator.enabled("broccoli-eyeglass:results") || debugGenerator.enabled("eyeglass:results");
        delete options.verbose;
        _this.renderSync = options.renderSync || false;
        delete options.renderSync;
        _this.discover = options.discover;
        delete options.discover;
        if (!options.sourceFiles) {
            _this.sourceFiles = [];
            if (_this.discover === false) {
                throw new Error("sourceFiles are required when discovery is disabled.");
            }
            else {
                // Default to discovery mode if no sourcefiles are provided.
                _this.discover = true;
            }
        }
        else {
            _this.sourceFiles = options.sourceFiles;
        }
        delete options.sourceFiles;
        _this.maxListeners = options.maxListeners || 10;
        delete options.maxListeners;
        _this.options = copyObject(options);
        if (!_this.cssDir) {
            throw new Error("Expected cssDir option.");
        }
        forbidNodeSassOption(_this.options, "file");
        forbidNodeSassOption(_this.options, "data");
        forbidNodeSassOption(_this.options, "outFile");
        if (_this.verbose) {
            _this.colors = require("colors/safe");
            _this.events.on("compiled", _this.logCompilationSuccess.bind(_this));
            _this.events.on("failed", _this.logCompilationFailure.bind(_this));
        }
        _this.events.addListener("compiled", function (details, result) {
            _this.addOutput(details.fullSassFilename, details.fullCssFilename);
            var depFiles = result.stats.includedFiles;
            _this.addDependency(details.fullSassFilename, details.fullSassFilename);
            for (var i = 0; i < depFiles.length; i++) {
                _this.addDependency(details.fullSassFilename, depFiles[i]);
            }
        });
        return _this;
    }
    BroccoliSassCompiler.prototype.logCompilationSuccess = function (details, result) {
        var timeInSeconds = result.stats.duration / 1000.0;
        if (timeInSeconds === 0) {
            timeInSeconds = 0.001; // nothing takes zero seconds.
        }
        var action = this.colors.inverse.green("compile (" + timeInSeconds + "s)");
        var message = this.scopedFileName(details.sassFilename) + " => " + details.cssFilename;
        // eslint-disable-next-line no-console
        console.log(action + " " + message);
    };
    BroccoliSassCompiler.prototype.logCompilationFailure = function (details, error) {
        var sassFilename = details.sassFilename;
        var action = this.colors.bgRed.white("error");
        var message = this.colors.red(error.message);
        var location = "Line " + error.line + ", Column " + error.column;
        if (error.file.substring(error.file.length - sassFilename.length) !== sassFilename) {
            location = location + " of " + error.file;
        }
        // eslint-disable-next-line no-console
        console.log(action + " " + sassFilename + " (" + location + "): " + message);
    };
    BroccoliSassCompiler.prototype.compileTree = function (srcPath, files, destDir, compilationTimer) {
        var _this = this;
        switch (files.length) {
            case 0:
                return RSVP.resolve();
            case 1:
                return RSVP.all(this.compileSassFile(srcPath, files[0], destDir, compilationTimer));
            default: {
                var worker_1 = queue.async.asyncify(function (file) {
                    return RSVP.all(_this.compileSassFile(srcPath, file, destDir, compilationTimer));
                });
                return concurrency.then(function (numConcurrentCalls) {
                    concurrencyDebug("Compiling files with a worker queue of size %d", numConcurrentCalls);
                    return RSVP.resolve(queue(worker_1, files, numConcurrentCalls));
                });
            }
        }
    };
    BroccoliSassCompiler.prototype.compileSassFile = function (srcPath, sassFilename, destDir, compilationTimer) {
        var _this = this;
        var sassOptions = copyObject(this.options);
        var fullSassFilename = path.join(srcPath, sassFilename);
        sassOptions.file = fullSassFilename;
        var parsedName = parsePath(sassFilename);
        if (this.sassDir && parsedName.dir.slice(0, this.sassDir.length) === this.sassDir) {
            parsedName.dir = parsedName.dir.slice(this.sassDir.length + 1);
        }
        parsedName.ext = ".css";
        parsedName.base = parsedName.name + ".css";
        var cssFileName = path.join(this.cssDir, formatPath(parsedName));
        var promises = [];
        this.optionsGenerator(sassFilename, cssFileName, sassOptions, function (resolvedCssFileName, resolvedOptions) {
            var details = {
                srcPath: srcPath,
                sassFilename: sassFilename,
                fullSassFilename: resolvedOptions.file || fullSassFilename,
                destDir: destDir,
                cssFilename: resolvedCssFileName,
                fullCssFilename: path.join(destDir, resolvedCssFileName),
                options: copyObject(resolvedOptions),
            };
            details.options.outFile = details.cssFilename;
            promises.push(_this.compileCssFileMaybe(details, compilationTimer));
        });
        return promises;
    };
    BroccoliSassCompiler.prototype.renderer = function () {
        initSass();
        if (this.renderSync) {
            return renderSassSync;
        }
        else {
            return renderSass;
        }
    };
    /* Check if a dependency's hash has changed.
     *
     * @argument srcDir The directory in which to resolve relative paths against.
     * @argument dep An array of two elements, the first is the file and
     *               the second is the last known hash of that file.
     *
     * @return Boolean true if the file hasn't changed from the input hash, otherwise false
     **/
    BroccoliSassCompiler.prototype.dependencyChanged = function (srcDir, dep) {
        var file = path.isAbsolute(dep[0]) ? dep[0] : path.join(srcDir, dep[0]);
        var hexDigest = dep[1];
        return hexDigest !== this.hashForFile(file);
    };
    /* get the cached output for a source file, or compile the file if not in cache
     *
     * @argument details The compilation details object.
     *
     * @return Promise that resolves to the cached output of the file or the output
     *   of compiling the file
     **/
    BroccoliSassCompiler.prototype.getFromCacheOrCompile = function (details, compilationTimer) {
        var _this = this;
        var key = this.keyForSourceFile(details.srcPath, details.sassFilename, details.options);
        try {
            var cachedDependencies = this.persistentCache.get(this.dependenciesKey(key));
            if (!cachedDependencies.isCached) {
                var reason = { message: "no dependency data for " + details.sassFilename };
                return this.handleCacheMiss(details, reason, key, compilationTimer);
            }
            var dependencies = JSON.parse(cachedDependencies.value);
            // check dependency caches
            if (dependencies.some(function (dep) { return _this.dependencyChanged(details.srcPath, dep); })) {
                var reason = { message: "dependency changed" };
                return this.handleCacheMiss(details, reason, key, compilationTimer);
            }
            var cachedOutput = this.persistentCache.get(this.outputKey(key));
            if (!cachedOutput.isCached) {
                var reason = { message: "no output data for " + details.sassFilename };
                return this.handleCacheMiss(details, reason, key, compilationTimer);
            }
            var depFiles = dependencies.map(function (depAndHash) { return depAndHash[0]; });
            var value = [depFiles, JSON.parse(cachedOutput.value)];
            compilationTimer.stats.cacheHitCount++;
            return RSVP.resolve(this.handleCacheHit(details, value).then(function () { }));
        }
        catch (error) {
            return this.handleCacheMiss(details, error, key, compilationTimer);
        }
    };
    /* compute the hash for a file.
     *
     * @argument absolutePath The absolute path to the file.
     * @return hash object of the file data
     **/
    BroccoliSassCompiler.prototype.hashForFile = function (absolutePath) {
        return this.fileKey(absolutePath);
    };
    /* compute a key for a file that will change if the file has changed. */
    BroccoliSassCompiler.prototype.fileKey = function (file) {
        var cachedKeyKey = "fileKey(" + file + ")";
        var cachedKey = this.buildCache.get(cachedKeyKey);
        if (cachedKey) {
            return cachedKey;
        }
        var key;
        try {
            var stat = fs_1.statSync(file);
            key = mtimeMs(stat) + ":" + stat.size + ":" + stat.mode;
            this.buildCache.set(cachedKeyKey, key);
        }
        catch (_) {
            key = "0:0:0";
        }
        this.buildCache.set(cachedKeyKey, key);
        return key;
    };
    /* construct a base cache key for a file to be compiled.
     *
     * @argument srcDir The directory in which to resolve relative paths against.
     * @argument relativeFilename The filename relative to the srcDir that is being compiled.
     * @argument options The compilation options.
     *
     * @return The cache key for the file
     **/
    BroccoliSassCompiler.prototype.keyForSourceFile = function (srcDir, relativeFilename, _options) {
        var absolutePath = path.join(srcDir, relativeFilename);
        var hash = this.hashForFile(absolutePath);
        return relativeFilename + "@" + hash;
    };
    /* construct a cache key for storing dependency hashes.
     *
     * @argument key The base cache key
     * @return String
     */
    BroccoliSassCompiler.prototype.dependenciesKey = function (key) {
        return "[[[dependencies of " + key + "]]]";
    };
    /* construct a cache key for storing output.
     *
     * @argument key The base cache key
     * @return String
     */
    BroccoliSassCompiler.prototype.outputKey = function (key) {
        return "[[[output of " + key + "] v2]]";
    };
    /* retrieve the files from cache, write them, and populate the hot cache information
     * for rebuilds.
     */
    BroccoliSassCompiler.prototype.handleCacheHit = function (details, inputAndOutputFiles) {
        var e_1, _a, e_2, _b;
        var _this = this;
        var _c = __read(inputAndOutputFiles, 2), inputFiles = _c[0], outputFiles = _c[1];
        this.persistentCacheDebug("Persistent cache hit for %s. Writing to: %s", details.sassFilename, details.fullCssFilename);
        if (this.verbose) {
            var action = this.colors.inverse.green("cached");
            var message = this.scopedFileName(details.sassFilename) + " => " + details.cssFilename;
            // eslint-disable-next-line no-console
            console.log(action + " " + message);
        }
        inputFiles.forEach(function (dep) {
            // populate the dependencies cache for rebuilds
            _this.addDependency(details.fullSassFilename, path.resolve(details.srcPath, dep));
        });
        var contents = outputFiles.contents, urls = outputFiles.urls;
        var files = Object.keys(contents);
        this.persistentCacheDebug("cached output files for %s are: %s", details.sassFilename, files.join(", "));
        try {
            for (var files_1 = __values(files), files_1_1 = files_1.next(); !files_1_1.done; files_1_1 = files_1.next()) {
                var file = files_1_1.value;
                var data = contents[file];
                var cachedFile = path.join(this.cachePath, file);
                var outputFile = path.join(this.outputPath, file);
                // populate the output cache for rebuilds
                this.addOutput(details.fullSassFilename, outputFile);
                writeDataToFile(cachedFile, outputFile, Buffer.from(data, "base64"));
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (files_1_1 && !files_1_1.done && (_a = files_1.return)) _a.call(files_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
        var eventPromises = [];
        var allUrls = Object.keys(urls);
        if (allUrls.length > 0) {
            this.persistentCacheDebug("firing 'cached-asset' for each asset url for %s: %s", details.sassFilename, allUrls.join(", "));
        }
        try {
            for (var allUrls_1 = __values(allUrls), allUrls_1_1 = allUrls_1.next(); !allUrls_1_1.done; allUrls_1_1 = allUrls_1.next()) {
                var url = allUrls_1_1.value;
                var sourceFile = urls[url];
                eventPromises.push(this.events.emit("cached-asset", sourceFile, url));
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (allUrls_1_1 && !allUrls_1_1.done && (_b = allUrls_1.return)) _b.call(allUrls_1);
            }
            finally { if (e_2) throw e_2.error; }
        }
        return RSVP.all(eventPromises);
    };
    BroccoliSassCompiler.prototype.scopedFileName = function (file) {
        file = this.relativize(file);
        if (this.treeName) {
            return this.treeName + "/" + file;
        }
        else {
            return file;
        }
    };
    BroccoliSassCompiler.prototype.relativize = function (file) {
        return removePathPrefix(this.inputPaths[0], [file])[0];
    };
    BroccoliSassCompiler.prototype.isOutputInTree = function (file) {
        if (path.isAbsolute(file)) {
            return file.startsWith(this.outputPath);
        }
        else {
            return true;
        }
    };
    BroccoliSassCompiler.prototype.relativizeOutput = function (file) {
        return removePathPrefix(this.outputPath, [file])[0];
    };
    BroccoliSassCompiler.prototype.relativizeAll = function (files) {
        return removePathPrefix(this.inputPaths[0], files);
    };
    BroccoliSassCompiler.prototype.hasDependenciesSet = function (file) {
        return this.dependencies[this.relativize(file)] !== undefined;
    };
    BroccoliSassCompiler.prototype.dependenciesOf = function (file) {
        return this.dependencies[this.relativize(file)] || new Set();
    };
    BroccoliSassCompiler.prototype.outputsFrom = function (file) {
        return this.outputs[this.relativize(file)] || new Set();
    };
    BroccoliSassCompiler.prototype.outputURLsFrom = function (file) {
        return this.outputURLs[this.relativize(file)] || new Map();
    };
    /**
     * Some filenames returned from importers are not really files
     * on disk. These three prefixes are used in eyeglass.
     * Skipping a read on these files avoids a more expensive fs call
     * and exception handling.
     * @param filename a filename returned from an importer
     */
    BroccoliSassCompiler.prototype.isNotFile = function (filename) {
        return filename.startsWith("already-imported:") ||
            filename.startsWith("autoGenerated:") ||
            filename.startsWith("fs:");
    };
    /* hash all dependencies synchronously and return the files that exist
     * as an array of pairs (filename, hash).
     */
    BroccoliSassCompiler.prototype.hashDependencies = function (details) {
        var _this = this;
        var depsWithHashes = new Array();
        this.dependenciesOf(details.fullSassFilename).forEach(function (f) {
            try {
                if (_this.isNotFile(f)) {
                    _this.persistentCacheDebug("Ignoring non-file dependency: %s", f);
                }
                else {
                    var h = _this.hashForFile(f);
                    if (f.startsWith(details.srcPath)) {
                        f = f.substring(details.srcPath.length + 1);
                    }
                    depsWithHashes.push([f, h]);
                }
            }
            catch (e) {
                if (typeof e === "object" && e !== null && e.code === "ENOENT") {
                    _this.persistentCacheDebug("Ignoring non-existent file: %s", f);
                }
                else {
                    throw e;
                }
            }
        });
        // prune out the dependencies that weren't files.
        return depsWithHashes;
    };
    /* read all output files asynchronously and return the contents
     * as a hash of relative filenames to base64 encoded strings.
     */
    BroccoliSassCompiler.prototype.readOutputs = function (details) {
        var e_3, _a, e_4, _b;
        var contents = {};
        var urls = {};
        var outputs = this.outputsFrom(details.fullSassFilename);
        try {
            for (var outputs_1 = __values(outputs), outputs_1_1 = outputs_1.next(); !outputs_1_1.done; outputs_1_1 = outputs_1.next()) {
                var output = outputs_1_1.value;
                if (this.isOutputInTree(output)) {
                    contents[this.relativizeOutput(output)] = fs.readFileSync(output, "base64");
                }
                else {
                    this.persistentCacheDebug("refusing to cache output file found outside the output tree: %s", output);
                }
            }
        }
        catch (e_3_1) { e_3 = { error: e_3_1 }; }
        finally {
            try {
                if (outputs_1_1 && !outputs_1_1.done && (_a = outputs_1.return)) _a.call(outputs_1);
            }
            finally { if (e_3) throw e_3.error; }
        }
        var outputURLs = this.outputURLsFrom(details.fullSassFilename);
        try {
            for (var _c = __values(outputURLs.keys()), _d = _c.next(); !_d.done; _d = _c.next()) {
                var url = _d.value;
                urls[url] = outputURLs.get(url);
            }
        }
        catch (e_4_1) { e_4 = { error: e_4_1 }; }
        finally {
            try {
                if (_d && !_d.done && (_b = _c.return)) _b.call(_c);
            }
            finally { if (e_4) throw e_4.error; }
        }
        return { contents: contents, urls: urls };
    };
    /* Writes the dependencies and output contents to the persistent cache */
    BroccoliSassCompiler.prototype.populateCache = function (key, details, _result) {
        this.persistentCacheDebug("Populating cache for " + key);
        var cache = this.persistentCache;
        var depsWithHashes = this.hashDependencies(details);
        var outputContents = this.readOutputs(details);
        cache.set(this.dependenciesKey(key), JSON.stringify(depsWithHashes));
        cache.set(this.outputKey(key), JSON.stringify(outputContents));
    };
    /* When the cache misses, we need to compile the file and then populate the cache */
    BroccoliSassCompiler.prototype.handleCacheMiss = function (details, reason, key, compilationTimer) {
        var _this = this;
        compilationTimer.stats.cacheMissCount++;
        this.persistentCacheDebug("Persistent cache miss for %s. Reason: %s", details.sassFilename, reason.message);
        // for errors
        if (reason.stack) {
            this.persistentCacheDebug("Stacktrace:", reason.stack);
        }
        return this.compileCssFile(details, compilationTimer).then(function (result) {
            return _this.populateCache(key, details, result);
        });
    };
    /* Compile the file if it's not in the cache.
     * Reuse cached output if it is.
     *
     * @argument details The compilation details object.
     *
     * @return A promise that resolves when the output files are written
     *   either from cache or by compiling. Rejects on error.
     */
    BroccoliSassCompiler.prototype.compileCssFileMaybe = function (details, compilationTimer) {
        if (this.persistentCache) {
            return this.getFromCacheOrCompile(details, compilationTimer);
        }
        else {
            return this.compileCssFile(details, compilationTimer);
        }
    };
    BroccoliSassCompiler.prototype.compileCssFile = function (details, compilationTimer) {
        var _this = this;
        var sass = this.renderer();
        var success = this.handleSuccess.bind(this, details);
        var failure = this.handleFailure.bind(this, details);
        return this.events.emit("compiling", details).then(function () {
            var dependencyListener = function (absolutePath) {
                _this.addDependency(details.fullSassFilename, absolutePath);
            };
            var additionalOutputListener = function (absolutePathToOutput, httpPathToOutput, absolutePathToSource) {
                _this.persistentCacheDebug("additional-output %s -> %s -> %s", absolutePathToSource, httpPathToOutput, absolutePathToOutput);
                if (!_this.isOutputInTree(absolutePathToOutput)) {
                    // it's outside this tree, don't cache the output.
                    if (absolutePathToSource && httpPathToOutput) {
                        _this.persistentCacheDebug("additional-output is outside tree will cache source & url");
                        // something outside this tree is putting it there, so we need to
                        // let that same thing deal with it again when the warm cache is accessed.
                        // we will track this file from its source location and target url
                        _this.addSource(details.fullSassFilename, absolutePathToSource, httpPathToOutput);
                    }
                }
                else {
                    _this.persistentCacheDebug("additional-output is in tree will cache contents");
                    _this.addOutput(details.fullSassFilename, absolutePathToOutput);
                }
            };
            _this.events.addListener("additional-output", additionalOutputListener);
            _this.events.addListener("dependency", dependencyListener);
            var sassPromise = sass(details.options); // XXX This cast sucks
            return sassPromise
                .finally(function () {
                _this.events.removeListener("dependency", dependencyListener);
                _this.events.removeListener("additional-output", additionalOutputListener);
            })
                .then(function (result) {
                var e_5, _a;
                compilationTimer.stats.nodeSassTime += result.stats.duration;
                compilationTimer.stats.importCount += result.stats.includedFiles.length;
                try {
                    for (var _b = __values(result.stats.includedFiles), _c = _b.next(); !_c.done; _c = _b.next()) {
                        var f = _c.value;
                        if (!f.startsWith("already-imported:")) {
                            compilationTimer.stats.uniqueImportCount++;
                        }
                    }
                }
                catch (e_5_1) { e_5 = { error: e_5_1 }; }
                finally {
                    try {
                        if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
                    }
                    finally { if (e_5) throw e_5.error; }
                }
                debug("render of " + result.stats.entry + " took " + result.stats.duration);
                return success(result).then(function () { return result; });
            }, failure);
        });
    };
    BroccoliSassCompiler.prototype.handleSuccess = function (details, result) {
        var cachedFile = path.join(this.cachePath, details.cssFilename);
        var outputFile = details.fullCssFilename;
        writeDataToFile(cachedFile, outputFile, result.css);
        return this.events.emit("compiled", details, result);
    };
    BroccoliSassCompiler.prototype.handleFailure = function (details, error) {
        var failed = this.events.emit("failed", details, error);
        return failed.then(function () {
            if (typeof error === "object" && error !== null) {
                error.message =
                    error.message + "\n    at " + error.file + ":" + error.line + ":" + error.column;
            }
            throw error;
        });
    };
    BroccoliSassCompiler.prototype.filesInTree = function (srcPath) {
        var sassDir = this.sassDir || "";
        var files = new Array();
        if (this.discover) {
            var pattern = path.join(srcPath, sassDir, "**", "[^_]*.scss");
            files = glob.sync(pattern);
        }
        this.sourceFiles.forEach(function (sourceFile) {
            var pattern = path.join(srcPath, sassDir, sourceFile);
            files = files.concat(glob.sync(pattern));
        });
        return unique(files);
    };
    BroccoliSassCompiler.prototype.addSource = function (sassFilename, sourceFilename, httpPathToOutput) {
        sassFilename = this.relativize(sassFilename);
        this.outputURLs[sassFilename] = this.outputURLs[sassFilename] || new Map();
        var urlMap = this.outputURLs[sassFilename];
        urlMap.set(httpPathToOutput, sourceFilename);
    };
    BroccoliSassCompiler.prototype.addOutput = function (sassFilename, outputFilename) {
        sassFilename = this.relativize(sassFilename);
        this.outputs[sassFilename] = this.outputs[sassFilename] || new Set();
        this.outputs[sassFilename].add(outputFilename);
    };
    BroccoliSassCompiler.prototype.clearOutputs = function (files) {
        var _this = this;
        this.relativizeAll(files).forEach(function (f) {
            if (_this.outputs[f]) {
                delete _this.outputs[f];
            }
            if (_this.outputURLs[f]) {
                delete _this.outputURLs[f];
            }
        });
    };
    /* This method computes the output files that are only output for at least one given inputs
     * and never for an input that isn't provided.
     *
     * This is important because the same assets might be output from compiling several
     * different inputs for tools like eyeglass assets.
     *
     * @return Set<String> The full paths output files.
     */
    BroccoliSassCompiler.prototype.outputsFromOnly = function (inputs) {
        inputs = this.relativizeAll(inputs);
        var otherOutputs = new Set();
        var onlyOutputs = new Set();
        var allInputs = Object.keys(this.outputs);
        for (var i = 0; i < allInputs.length; i++) {
            var outputs = this.outputs[allInputs[i]];
            if (inputs.indexOf(allInputs[i]) < 0) {
                outputs.forEach(function (output) { return otherOutputs.add(output); });
            }
            else {
                outputs.forEach(function (output) { return onlyOutputs.add(output); });
            }
        }
        onlyOutputs.forEach(function (only) {
            if (otherOutputs.has(only)) {
                onlyOutputs.delete(only);
            }
        });
        return onlyOutputs;
    };
    BroccoliSassCompiler.prototype.addDependency = function (sassFilename, dependencyFilename) {
        sassFilename = this.relativize(sassFilename);
        this.dependencies[sassFilename] = this.dependencies[sassFilename] || new Set();
        this.dependencies[sassFilename].add(dependencyFilename);
    };
    BroccoliSassCompiler.prototype.clearDependencies = function (files) {
        var _this = this;
        this.relativizeAll(files).forEach(function (f) {
            delete _this.dependencies[f];
        });
    };
    BroccoliSassCompiler.prototype.knownDependencies = function () {
        var deps = new Set();
        var sassFiles = Object.keys(this.dependencies);
        for (var i = 0; i < sassFiles.length; i++) {
            var sassFile = sassFiles[i];
            deps.add(sassFile);
            this.dependencies[sassFile].forEach(function (dep) { return deps.add(dep); });
        }
        var entries = new Array();
        deps.forEach(function (d) {
            try {
                entries.push(new Entry(d));
            }
            catch (e) {
                // Lots of things aren't files that are dependencies, ignore them.
            }
        });
        return entries;
    };
    BroccoliSassCompiler.prototype.hasKnownDependencies = function () {
        return Object.keys(this.dependencies).length > 0;
    };
    BroccoliSassCompiler.prototype.knownDependenciesTree = function (inputPath) {
        var entries = walkSync.entries(inputPath);
        absolutizeEntries(entries);
        var tree = FSTreeFromEntries(entries);
        tree.addEntries(this.knownDependencies(), { sortAndExpand: true });
        return tree;
    };
    BroccoliSassCompiler.prototype._reset = function () {
        this.currentTree = null;
        this.dependencies = {};
        this.outputs = {};
        this.outputURLs = {};
    };
    BroccoliSassCompiler.prototype._build = function () {
        var _this = this;
        var inputPath = this.inputPaths[0];
        var outputPath = this.outputPath;
        var currentTree = this.currentTree;
        var nextTree = null;
        var patches = new Array();
        var compilationAvoidanceTimer = heimdall.start("eyeglass:broccoli:build:invalidation");
        if (this.hasKnownDependencies()) {
            hotCacheDebug("Has known dependencies");
            nextTree = this.knownDependenciesTree(inputPath);
            this.currentTree = nextTree;
            currentTree = currentTree || new FSTree();
            patches = currentTree.calculatePatch(nextTree);
            hotCacheDebug("currentTree = ", currentTree);
            hotCacheDebug("nextTree = ", nextTree);
            hotCacheDebug("patches = ", patches);
        }
        else {
            hotCacheDebug("No known dependencies");
        }
        // TODO: handle indented syntax files.
        var treeFiles = removePathPrefix(inputPath, this.filesInTree(inputPath));
        treeFiles = treeFiles.filter(function (f) {
            f = path.join(inputPath, f);
            if (!_this.hasDependenciesSet(f)) {
                hotCacheDebug("no deps for", _this.scopedFileName(f));
                return true;
            }
            var deps = _this.dependenciesOf(f);
            hotCacheDebug("dependencies are", deps);
            for (var p = 0; p < patches.length; p++) {
                var entry = patches[p][2];
                hotCacheDebug("looking for", entry.relativePath);
                if (deps.has(entry.relativePath)) {
                    hotCacheDebug("building because", entry.relativePath, "is used by", f);
                    return true;
                }
            }
            if (_this.verbose) {
                var action = _this.colors.inverse.green("unchanged");
                var message = _this.scopedFileName(f);
                // eslint-disable-next-line no-console
                console.log(action + " " + message);
            }
            return false;
        });
        compilationAvoidanceTimer.stop();
        // Cleanup any unneeded output files
        var removed = [];
        for (var p = 0; p < patches.length; p++) {
            if (patches[p][0] === "unlink") {
                var entry = patches[p][2];
                if (entry.relativePath.indexOf(inputPath) === 0) {
                    removed.push(entry.relativePath);
                }
            }
        }
        if (removed.length > 0) {
            var outputs = this.outputsFromOnly(removed);
            // TODO: outputURLsFromOnly(removed)
            outputs.forEach(function (output) {
                if (output.indexOf(outputPath) === 0) {
                    fs.unlinkSync(output);
                }
                else {
                    hotCacheDebug("not removing because outside the outputTree", output);
                    _this.events.emit("stale-external-output", output);
                }
            });
            this.clearOutputs(removed);
        }
        hotCacheDebug("building files:", treeFiles);
        var absoluteTreeFiles = treeFiles.map(function (f) { return path.join(inputPath, f); });
        this.clearDependencies(absoluteTreeFiles);
        this.clearOutputs(absoluteTreeFiles);
        var internalListeners = absoluteTreeFiles.length * 2 + // 1 dep & 1 output listeners each
            1 + // one compilation listener
            (this.verbose ? 2 : 0); // 2 logging listeners if in verbose mode
        debug("There are %d internal event listeners.", internalListeners);
        debug("Setting max external listeners to %d via the maxListeners option (default: 10).", this.maxListeners);
        this.events.setMaxListeners(internalListeners + this.maxListeners);
        var compilationTimer = heimdall.start("eyeglass:broccoli:build:compileTree:" + inputPath, SassRenderSchema);
        compilationTimer.stats.numSassFiles = treeFiles.length;
        return this.compileTree(inputPath, treeFiles, outputPath, compilationTimer).finally(function () {
            compilationTimer.stop();
            if (!_this.currentTree) {
                _this.currentTree = _this.knownDependenciesTree(inputPath);
            }
        });
    };
    BroccoliSassCompiler.prototype.build = function () {
        var _this = this;
        this.buildCount++;
        if (this.buildCount > 1) {
            this.buildCache = this.sessionCache || new Map();
        }
        else {
            if (process.env.BROCCOLI_EYEGLASS === "forceInvalidateCache") {
                this.persistentCacheDebug("clearing cache because forceInvalidateCache was set.");
                this.persistentCache && this.persistentCache.clear();
            }
        }
        return this._build().catch(function (e) {
            _this._reset();
            fs.removeSync(_this.outputPath);
            fs.mkdirpSync(_this.outputPath);
            throw e;
        });
    };
    return BroccoliSassCompiler;
}(BroccoliPlugin));
exports.default = BroccoliSassCompiler;
var SassRenderSchema = /** @class */ (function () {
    function SassRenderSchema() {
        this.cacheMissCount = 0;
        this.cacheHitCount = 0;
        this.numSassFiles = 0;
        this.nodeSassTime = 0;
        this.importCount = 0;
        this.uniqueImportCount = 0;
    }
    return SassRenderSchema;
}());
module.exports.shouldPersist = shouldPersist;
/* shim for fs.Stats.mtimeMS which was introduced in node 8. */
function mtimeMs(stat) {
    if (stat.mtimeMs) {
        return stat.mtimeMs;
    }
    else {
        return stat.mtime.valueOf();
    }
}
//# sourceMappingURL=broccoli_sass_compiler.js.map